@@ -14,20 +14,31 @@
 .ENSEMBL_FTP_PUB_GRCH37_URL <- "ftp://ftp.ensembl.org/pub/grch37/"
 .ENSEMBLGENOMES_FTP_PUB_URL <- "ftp://ftp.ensemblgenomes.org/pub/"
 .ENSEMBL_FTP_RELEASE_PREFIX <- "release-"
+.ENSEMBL_DIVISIONS <- c("bacteria", "fungi", "metazoa", "plants", "protists")
 
-### 'division' must be NA or one of the Ensembl Genomes divisions i.e.
-### "bacteria", "fungi", "metazoa", "plants", or "protists".
+.normarg_division <- function(division, use.grch37=FALSE)
+{
+    if (!isSingleStringOrNA(division))
+        stop(wmsg("'division' must be a single string (or NA)"))
+    if (is.na(division))
+        return(NA)
+    if (division == "")
+        stop(wmsg("'division' cannot be the empty string"))
+    if (use.grch37)
+        stop(wmsg("'division' and 'use.grch37' cannot both be specified"))
+    if (!(division %in% .ENSEMBL_DIVISIONS))
+        stop(wmsg("'division' should be one of ",
+                  paste0('"', .ENSEMBL_DIVISIONS, '"', collapse=", ")))
+    division
+}
+
+### 'division' must be NA or one of the Ensembl Genomes divisions.
 .get_Ensembl_FTP_top_url <- function(division=NA, use.grch37=FALSE)
 {
-    if (!is_single_value(division))
-        stop(wmsg("'division' must be a single value"))
     if (!isTRUEorFALSE(use.grch37))
         stop(wmsg("'use.grch37' must be TRUE or FALSE"))
+    division <- .normarg_division(division, use.grch37)
     if (!is.na(division)) {
-        if (!is.character(division) || division == "")
-            stop(wmsg("'division' must be a single non-empty string or NA"))
-        if (use.grch37)
-            stop(wmsg("'division' and 'use.grch37' cannot both be specified"))
         top_url <- paste0(.ENSEMBLGENOMES_FTP_PUB_URL, division, "/")
     } else if (use.grch37) {
         top_url <- .ENSEMBL_FTP_PUB_GRCH37_URL
@@ -37,15 +48,13 @@
     top_url
 }
 
-### 'division' must be NA or one of the Ensembl Genomes divisions i.e.
-### "bacteria", "fungi", "metazoa", "plants", or "protists".
+### 'division' must be NA or one of the Ensembl Genomes divisions.
 get_Ensembl_FTP_mysql_url <- function(release=NA, division=NA,
                                       use.grch37=FALSE)
 {
     if (!is_single_value(release))
         stop(wmsg("'release' must be a single value"))
-    top_url <- .get_Ensembl_FTP_top_url(division=division,
-                                        use.grch37=use.grch37)
+    top_url <- .get_Ensembl_FTP_top_url(division, use.grch37)
     if (!is.na(release)) {
         mysql_subdir <- paste0(.ENSEMBL_FTP_RELEASE_PREFIX, release, "/mysql")
     } else if (is.na(division) && !use.grch37) {
@@ -56,15 +65,13 @@ get_Ensembl_FTP_mysql_url <- function(release=NA, division=NA,
     paste0(top_url, mysql_subdir, "/")
 }
 
-### 'division' must be NA or one of the Ensembl Genomes divisions i.e.
-### "bacteria", "fungi", "metazoa", "plants", or "protists".
+### 'division' must be NA or one of the Ensembl Genomes divisions.
 get_Ensembl_FTP_gtf_url <- function(release=NA, division=NA,
                                     use.grch37=FALSE)
 {
     if (!is_single_value(release))
         stop(wmsg("'release' must be a single value"))
-    top_url <- .get_Ensembl_FTP_top_url(division=division,
-                                        use.grch37=use.grch37)
+    top_url <- .get_Ensembl_FTP_top_url(division, use.grch37)
     if (!is.na(release)) {
         gtf_subdir <- paste0(.ENSEMBL_FTP_RELEASE_PREFIX, release, "/gtf")
     } else if (is.na(division) && !use.grch37) {
@@ -76,6 +83,35 @@ get_Ensembl_FTP_gtf_url <- function(release=NA, division=NA,
 }
 
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### .list_Ensembl_FTP_core_dbs()
+###
+
+### The keys are FTP URLs to "mysql" directories e.g.
+###   "ftp://ftp.ensembl.org/pub/current_mysql/"
+###   "ftp://ftp.ensembl.org/pub/release-98/mysql/"
+###   "ftp://ftp.ensemblgenomes.org/pub/bacteria/current/mysql/"
+###   "ftp://ftp.ensemblgenomes.org/pub/plants/release-45/mysql/"
+### etc...
+.Ensembl_FTP_cached_core_dbs <- new.env(parent=emptyenv())
+
+.list_Ensembl_FTP_core_dbs <- function(mysql_url, release=NA)
+{
+    stopifnot(isSingleString(mysql_url))
+    pattern <- "_core_"
+    core_dbs <- .Ensembl_FTP_cached_core_dbs[[mysql_url]]
+    if (is.null(core_dbs)) {
+        subdirs <- list_ftp_dir(mysql_url, subdirs.only=TRUE)
+        core_dbs <- subdirs[grep(pattern, subdirs, fixed=TRUE)]
+        .Ensembl_FTP_cached_core_dbs[[mysql_url]] <- core_dbs
+    }
+    if (is.na(release))
+        return(core_dbs)
+    pattern <- paste0(pattern, release, "_")
+    core_dbs[grep(pattern, core_dbs, fixed=TRUE)]
+}
+
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### get_current_Ensembl_release()
 ###
@@ -87,13 +123,13 @@ get_Ensembl_FTP_gtf_url <- function(release=NA, division=NA,
 ### etc...
 .Ensembl_FTP_cached_releases <- new.env(parent=emptyenv())
 
+### 'division' must be NA or one of the Ensembl Genomes divisions.
 .list_Ensembl_FTP_releases <- function(division=NA, use.grch37=FALSE,
                                        as.subdirs=FALSE)
 {
     if (!isTRUEorFALSE(as.subdirs))
         stop(wmsg("'as.subdirs' must be TRUE or FALSE"))
-    top_url <- .get_Ensembl_FTP_top_url(division=division,
-                                        use.grch37=use.grch37)
+    top_url <- .get_Ensembl_FTP_top_url(division, use.grch37)
     releases <- .Ensembl_FTP_cached_releases[[top_url]]
     if (is.null(releases)) {
         top_files <- list_ftp_dir(top_url)
@@ -109,22 +145,33 @@ get_Ensembl_FTP_gtf_url <- function(release=NA, division=NA,
     releases
 }
 
+### 'division' must be NA or one of the Ensembl Genomes divisions.
 get_current_Ensembl_release <- function(division=NA, use.grch37=FALSE)
 {
-    releases <- .list_Ensembl_FTP_releases(division=division,
-                                           use.grch37=use.grch37)
+    releases <- .list_Ensembl_FTP_releases(division, use.grch37)
     current_release <- max(as.integer(releases))
-    ## The latest release is not necessarily **officially** released, in
-    ## which case the corresponding FTP dir is incomplete. The following
-    ## code tries to detect this situation by checking the presence of the
-    ## README file.
-    top_url <- .get_Ensembl_FTP_top_url(division=division,
-                                        use.grch37=use.grch37)
-    README_url <- paste0(top_url, .ENSEMBL_FTP_RELEASE_PREFIX,
-                         current_release, "/README")
-    doc <- try(RCurl::getURL(README_url), silent=TRUE)
-    if (inherits(doc, "try-error"))
-        current_release <- current_release - 1L
+    ## We should be done here. However, sometimes the Ensembl folks create
+    ## FTP folder ftp.ensembl.org/pub/release-<current_release>/ a few days
+    ## before it's officially released. In this case the folder typically
+    ## misses a bunch of subfolders like mysql/. So we check that subfolder
+    ## mysql/ can be listed and contains the right thing.
+    while (TRUE) {
+        mysql_url <- get_Ensembl_FTP_mysql_url(current_release, division,
+                                               use.grch37)
+        core_dbs <- try(suppressWarnings(
+                            .list_Ensembl_FTP_core_dbs(mysql_url)
+                        ), silent=TRUE)
+        if (inherits(core_dbs, "try-error") || length(core_dbs) == 0L) {
+            current_release <- current_release - 1L
+            next
+        }
+        ok <- grepl(paste0("_core_", current_release, "_"), core_dbs)
+        if (!all(ok)) {
+            current_release <- current_release - 1L
+            next
+        }
+        break
+    }
     current_release
 }
 
@@ -140,15 +187,11 @@ use_species_index_from_Ensembl_FTP <- function(release=NA, division=NA,
 {
     if (!is_single_value(release))
         stop(wmsg("'release' must be a single value"))
-    if (!is_single_value(division))
-        stop(wmsg("'division' must be a single value"))
     if (!isTRUEorFALSE(use.grch37))
         stop(wmsg("'use.grch37' must be TRUE or FALSE"))
-    if (use.grch37) {
-        if (is.na(division))
-            return(FALSE)
-        stop(wmsg("'division' and 'use.grch37' cannot both be specified"))
-    }
+    division <- .normarg_division(division, use.grch37)
+    if (use.grch37)
+        return(FALSE)
     if (is.na(release))
         return(TRUE)
     release <- as.integer(release)
@@ -161,7 +204,7 @@ use_species_index_from_Ensembl_FTP <- function(release=NA, division=NA,
 {
     if (!is_single_value(release))
         stop(wmsg("'release' must be a single value"))
-    top_url <- .get_Ensembl_FTP_top_url(division=division)
+    top_url <- .get_Ensembl_FTP_top_url(division)
     if (is.na(division)) {
         ## Available in Ensembl release 96 (March 2019) and above.
         species_file <- "species_EnsemblVertebrates.txt"
@@ -367,11 +410,11 @@ fetch_species_index_from_Ensembl_FTP <- function(release=NA, division=NA)
 ###
 
 .extract_species_info_from_species_index <- function(species_index, idx,
-                                                     release=NA)
+                                                     release=NA, division=NA)
 {
     stopifnot(is_single_value(release))
     if (is.na(release)) {
-        release <- get_current_Ensembl_release()
+        release <- get_current_Ensembl_release(division)
     } else {
         release <- as.integer(release)
     }
@@ -391,7 +434,7 @@ fetch_species_index_from_Ensembl_FTP <- function(release=NA, division=NA)
 {
     stopifnot(is_single_value(release))
     if (is.na(release)) {
-        release <- get_current_Ensembl_release()
+        release <- get_current_Ensembl_release(use.grch37=TRUE)
     } else {
         release <- as.integer(release)
     }
@@ -432,7 +475,8 @@ check_species_info <- function(species_info)
     species_index <- .load_or_fetch_species_index_from_url(url)
     idx <- .lookup_species_in_species_index(species, species_index, url)
     species_info <- .extract_species_info_from_species_index(species_index,
-                                                             idx, release)
+                                                             idx,
+                                                             release, division)
     core_db <- species_info$core_db
     attr(core_db, "species_info") <- species_info
     core_db
@@ -469,35 +513,10 @@ check_species_info <- function(species_info)
     core_db
 }
 
-### The keys are FTP URLs to "mysql" directories e.g.
-###   "ftp://ftp.ensembl.org/pub/current_mysql/"
-###   "ftp://ftp.ensembl.org/pub/release-98/mysql/"
-###   "ftp://ftp.ensemblgenomes.org/pub/bacteria/current/mysql/"
-###   "ftp://ftp.ensemblgenomes.org/pub/plants/release-45/mysql/"
-### etc...
-.Ensembl_FTP_cached_core_dbs <- new.env(parent=emptyenv())
-
-.list_Ensembl_FTP_core_dbs <- function(mysql_url, release=NA)
-{
-    stopifnot(isSingleString(mysql_url))
-
-    pattern <- "_core_"
-    core_dbs <- .Ensembl_FTP_cached_core_dbs[[mysql_url]]
-    if (is.null(core_dbs)) {
-        subdirs <- list_ftp_dir(mysql_url, subdirs.only=TRUE)
-        core_dbs <- subdirs[grep(pattern, subdirs, fixed=TRUE)]
-        .Ensembl_FTP_cached_core_dbs[[mysql_url]] <- core_dbs
-    }
-
-    if (!is.na(release))
-        pattern <- paste0(pattern, release, "_")
-    core_dbs[grep(pattern, core_dbs, fixed=TRUE)]
-}
-
 ### Returns a single string with the "species_info" attribute (named list)
 ### on it.
 .find_core_db_in_Ensembl_FTP_core_dbs <- function(mysql_url, species,
-                                                  release=NA)
+                                                  release=NA, division=NA)
 {
     stopifnot(isSingleString(species))
     core_dbs <- .list_Ensembl_FTP_core_dbs(mysql_url, release=release)
@@ -508,7 +527,7 @@ check_species_info <- function(species_info)
                   "\" species at ", mysql_url))
     core_db <- core_dbs[[idx]]
     if (is.na(release)) {
-        release <- get_current_Ensembl_release()
+        release <- get_current_Ensembl_release(division)
     } else {
         release <- as.integer(release)
     }
@@ -551,7 +570,8 @@ get_Ensembl_FTP_core_db_url <- function(species, release=NA, division=NA,
             ## The old way. Dumb, unreliable, and slow!
             core_db <- .find_core_db_in_Ensembl_FTP_core_dbs(
                                            mysql_url,
-                                           species, release=release)
+                                           species,
+                                           release=release, division=division)
         }
     }
     species_info <- attr(core_db, "species_info")
@@ -568,7 +588,8 @@ get_Ensembl_FTP_core_db_url <- function(species, release=NA, division=NA,
 
 fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
 {
-    coord_system <- fetch_table_from_Ensembl_FTP(core_db_url, "coord_system")
+    coord_system <- fetch_table_dump_from_Ensembl_FTP(core_db_url,
+                                                      "coord_system")
 
     ## Drop rows that do not have the default_version attrib.
     keep_idx <- grep("default_version", coord_system[ , "attrib"], fixed=TRUE)
@@ -595,12 +616,13 @@ fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
 {
     stopifnot(is_primary_key(seq_region_ids))
 
-    all_synonyms <- fetch_table_from_Ensembl_FTP(core_db_url,
-                                                 "seq_region_synonym")
+    all_synonyms <- fetch_table_dump_from_Ensembl_FTP(core_db_url,
+                                                      "seq_region_synonym")
     keep_idx <- which(all_synonyms[ , "seq_region_id"] %in% seq_region_ids)
     synonyms <- S4Vectors:::extract_data_frame_rows(all_synonyms, keep_idx)
 
-    external_dbs <- fetch_table_from_Ensembl_FTP(core_db_url, "external_db")
+    external_dbs <- fetch_table_dump_from_Ensembl_FTP(core_db_url,
+                                                      "external_db")
     synonyms <- join_dfs(synonyms, external_dbs,
                          "external_db_id", "external_db_id",
                          keep.Rcol=TRUE)  # we'll drop it below
@@ -642,7 +664,7 @@ fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
     ##     core_db_url <- GenomeInfoDb:::get_Ensembl_FTP_core_db_url(
     ##                                               "hsapiens",
     ##                                               release=release)
-    ##     attrib_type <- GenomeInfoDb:::fetch_table_from_Ensembl_FTP(
+    ##     attrib_type <- GenomeInfoDb:::fetch_table_dump_from_Ensembl_FTP(
     ##                                               core_db_url,
     ##                                               "attrib_type")
     ##   }
@@ -650,9 +672,9 @@ fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
     ##   http://ftp.ensembl.org/pub/release-89/mysql/homo_sapiens_core_89_38/
     ## e.g. many of them (including dump file for "attrib_type" table) have
     ## extension .gz.bz2 which is a mess!
-    #attrib_type <- fetch_table_from_Ensembl_FTP(core_db_url, "attrib_type",
-    #                                            nrows=99L)
-    attrib_type <- fetch_table_from_Ensembl_FTP(core_db_url, "attrib_type")
+    #attrib_type <- fetch_table_dump_from_Ensembl_FTP(core_db_url,
+    #                                                 "attrib_type", nrows=99L)
+    attrib_type <- fetch_table_dump_from_Ensembl_FTP(core_db_url, "attrib_type")
 
     m <- solid_match(codes, attrib_type[ , "code"],
                      x_what="supplied code",
@@ -671,7 +693,7 @@ fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
         stop(wmsg("'db_names' must be a character vector"))
     if (anyDuplicated(db_names))
         stop(wmsg("'db_names' cannot contain duplicates"))
-    external_db <- fetch_table_from_Ensembl_FTP(core_db_url, "external_db")
+    external_db <- fetch_table_dump_from_Ensembl_FTP(core_db_url, "external_db")
     m <- solid_match(db_names, external_db[ , "db_name"],
                      x_what="supplied db_name",
                      table_what="\"external_db.db_name\" value")
@@ -688,8 +710,8 @@ fetch_default_coord_systems_from_Ensembl_FTP <- function(core_db_url)
 .fetch_attribs_from_Ensembl_FTP <- function(core_db_url, seq_region_ids,
                                             toplevel=FALSE, non_ref=FALSE)
 {
-    all_attribs <- fetch_table_from_Ensembl_FTP(core_db_url,
-                                                "seq_region_attrib")
+    all_attribs <- fetch_table_dump_from_Ensembl_FTP(core_db_url,
+                                                     "seq_region_attrib")
     attrib_type_id <- all_attribs[ , "attrib_type_id"]
     codes <- c("toplevel", "non_ref")
     code2id <- .attrib_type_codes_to_ids_from_Ensembl_FTP(core_db_url, codes)
@@ -724,7 +746,7 @@ fetch_seq_regions_from_Ensembl_FTP <- function(core_db_url,
     coord_systems <- fetch_default_coord_systems_from_Ensembl_FTP(core_db_url)
 
     ## Fetch table "seq_region".
-    seq_regions <- fetch_table_from_Ensembl_FTP(core_db_url, "seq_region")
+    seq_regions <- fetch_table_dump_from_Ensembl_FTP(core_db_url, "seq_region")
 
     ## INNER JOIN table "seq_region" with table "coord_system".
     Rtable <- "coord_system"
